int select_rr_weight (struct multipath * mp);
int select_pgfailback (struct multipath * mp);
int select_pgpolicy (struct multipath * mp);
int select_selector (struct multipath * mp);
int select_alias (struct multipath * mp);
int select_features (struct multipath * mp);
int select_hwhandler (struct multipath * mp);
int select_checker(struct path *pp);
int select_getuid (struct path * pp);
int select_prio (struct path * pp);
int select_no_path_retry(struct multipath *mp);
int select_flush_on_last_del(struct multipath *mp);
int select_minio(struct multipath *mp);
int select_mode(struct multipath *mp);
int select_uid(struct multipath *mp);
int select_gid(struct multipath *mp);
int select_fast_io_fail(struct multipath *mp);
int select_dev_loss(struct multipath *mp);
int select_reservation_key(struct multipath *mp);
int select_retain_hwhandler (struct multipath * mp);
int select_detect_prio(struct path * pp);
int select_deferred_remove(struct multipath *mp);
int select_delay_watch_checks (struct multipath * mp);
int select_delay_wait_checks (struct multipath * mp);
